from pathlib					import Path
from typing						import List
from pygwarts.magical.spells	import patronus
from pygwarts.hagrid.bloom		import Bloom
from pygwarts.hagrid.thrivables	import Tree








class Rejuvenation(Bloom):

	"""
		Implementation of Bloom object that processes only files by "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.
		The files processing (leafs planting) for current Bloom object __call__ relies on mandatory
		"grow" callable, which will be invoked for every single Path object file in "leafs" list,
		which declaration must be reachable for current Tree object "origin" by escalation and which
		must implement "planting interface" to accept following arguments:
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current "planting" source file;
			bough	- Path object that represents current "planting" destination parent folder.
		This Bloom object designed to be the dispatcher for file-copying operations, and name "Rejuvenation"
		based on this idea. The idea of separation of copying and moving operations is flexibility
		and security. One can use moving "planting" along with this object, but it must be noted, that
		the logging information that emitted by this object will reflect the process of file copying,
		as was originally designed.
		Any Exception raised by "grow" might not be handled by "grow", as originally designed, but
		will be handled by this Bloom object. In this case, after the Exception will be handled,
		the iteration over leafs list will be continued. The iteration will be stopped if either
		"origin" argument is not a Tree object, or "origin" is a Tree object but with invalid
		"bough" field, or the "grow" implementation either not callable or not found.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, twigs :List[Path], leafs :List[Path]):


		if	isinstance(leafs, list) and len(leafs):
			if	isinstance(origin, Tree) and isinstance((bough := getattr(origin, "bough", None)), Path):
				self.loggy.debug(f"Commencing rejuvenation of {len(leafs)} \"{str(bough)}\" leafs")


				# IMPORTANT: "grow" to be invoked from "origin" to ensure right escalation, as "planting"
				# invocation must be issued not from Bloom object but from Tree object to be processed.
				if	callable(grow := getattr(origin, "grow", None)):
					for leaf in leafs:


						try:

							# Actual "planting" invocation relies on callable "grow", that escalated from
							# the current Tree object, to implement "planting interface". The chain of calls,
							# that leads to this moment assumes all arguments are valid, cause the original
							# design of "planting" object doesn't involve arguments validation. That means
							# only "origin" validation at the beginning guards possible Exceptions, that might
							# be raised in "grow" because of arguments inconsistency, and which all won't be
							# handled by original "planting" design but will be caught back in current Bloom.
							grow(origin, sprout, branch, leaf, bough)
							self.mark_tree(str(origin), "grow")
							self.loggy.debug(f"Rejuvenated \"{leaf}\"")


						except	Exception as E:

							self.loggy.warning(f"Failed to rejuve leaf \"{leaf}\" due to {patronus(E)}")
				else:		self.announce_implement(str(origin), "grow")
			else:			self.announce_bough(str(origin), "rejuve")
		else:				self.loggy.debug(f"Branch \"{str(branch)}\" has no leafs to rejuve")








class Transfer(Bloom):

	"""
		Implementation of Bloom object that processes only files by "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.
		The files processing (leafs planting) for current Bloom object __call__ relies on mandatory
		"graft" callable, which will be invoked for every single Path object file in "leafs" list,
		which declaration must be reachable for current Tree object "origin" by escalation and which
		must implement "planting interface" to accept following arguments:
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current "planting" source file;
			bough	- Path object that represents current "planting" destination parent folder.
		This Bloom object designed to be the dispatcher for file-moving operations, and name "Transfer"
		based on this idea. The idea of separation of copying and moving operations is flexibility
		and security. One can use copying "planting" along with this object, but it must be noted, that
		the logging information that emitted by this object will reflect the process of file moving,
		as was originally designed.
		Any Exception raised by "graft" might not be handled by "graft", as originally designed, but
		will be handled by this Bloom object. In this case, after the Exception will be handled,
		the iteration over leafs list will be continued. The iteration will be stopped if either
		"origin" argument is not a Tree object, or "origin" is a Tree object but with invalid
		"bough" field, or the "graft" implementation either not callable or not found.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :str, twigs :List[str], leafs :List[str]):


		if	isinstance(leafs, list) and len(leafs):
			if	isinstance(origin, Tree) and isinstance((bough := getattr(origin, "bough", None)), Path):
				self.loggy.debug(f"Commencing transfer of {len(leafs)} \"{bough}\" leafs")


				# IMPORTANT: "graft" to be invoked from "origin" to ensure right escalation, as "planting"
				# invocation must be issued not from Bloom object but from Tree object to be processed.
				if	callable(graft := getattr(origin, "graft", None)):
					for leaf in leafs:


						try:

							# Actual "planting" invocation relies on callable "graft", that escalated from
							# the current Tree object, to implement "planting interface". The chain of calls,
							# that leads to this moment assumes all arguments are valid, cause the original
							# design of "planting" object doesn't involve arguments validation. That means
							# only "origin" validation at the beginning guards possible Exceptions, that might
							# be raised in "graft" because of arguments inconsistency, and which all won't be
							# handled by original "planting" design but will be caught back in current Bloom.
							graft(origin, sprout, branch, leaf, bough)
							self.mark_tree(str(origin), "graft")
							self.loggy.debug(f"Transferred \"{leaf}\"")


						except	Exception as E:

							self.loggy.warning(f"Failed to transfer leaf \"{leaf}\" due to {patronus(E)}")
				else:		self.announce_implement(str(origin), "graft")
			else:			self.announce_bough(str(origin), "transfer")
		else:				self.loggy.debug(f"Branch \"{str(branch)}\" has no leafs to transfer")










from typing												import Optional
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.irma.access.volume						import LibraryVolume
from pygwarts.irma.shelve								import LibraryShelf
from pygwarts.irma.shelve.casing						import shelf_case








class InducerCase(ControlledTransmutation):

	"""
		irma.shelve.casing utility decorator variant, that integrates the core "casing" functionality into
		mutable chain, by maintaining AccessInducer output. As ControlledTransmutation class, accepts
		positional argument "link", which must be a string that represents the name of LibraryShelf
		attribute, and optional keyword-only arguments "prep" and "post", which must be "casing" handling
		callables. The key for LibraryShelf will be a string representation of current mutable chain point,
		and it will be passed along with this three accepted arguments to "shelf_case" function. In
		mutable chain acts as a mutation - takes decorated Transmutable, which must be an AccessInducer
		class which accepts LibraryVolume object as argument to be used to obtain AccessInducer original
		output, along with maintaining a LibraryVolume mapping in LibraryShelf, where current "key" will be
		mapped with "shelf_case" result value. If "shelf_case" return value is not None it will be returned.
		Otherwise returns original AccessInducer output. Despite AccessInducer must always must return
		string, this decoration doesn't restrict "post" return value, but suggests it will either return
		a string or will be passed to another object, which will return string.
	"""

	def __init__(
					self,
					link	:str,
					*,
					prep	:Optional[Callable[[str],str]]		=(lambda T	: T),
					post	:Optional[Callable[[str,str],str]]	=(lambda T,P: T),
				):

		self.link = link
		self.prep = prep
		self.post = post

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		link = self.link
		post = self.post
		prep = self.prep


		class InducerCasing(geminio(layer)):
			def __call__(self, volume :LibraryVolume) -> str :


				if	(induce := super().__call__(volume)) is not None:
					if	isinstance(mapping := getattr(self, str(link), None), LibraryShelf | dict):


						if	str(volume) not in mapping : mapping[str(volume)] = dict()
						if	(casing := shelf_case(

							induce,
							key=str(self),
							shelf=mapping[str(volume)],
							prep=prep,
							post=post,

						))	is not None:

							self.loggy.debug(f"Obtained induce casing value \"{casing}\"")
							return casing


					self.loggy.debug(f"Failed to obtain induce casing value")
					return induce


		return	InducerCasing










from os										import access	as osaccess
from os										import W_OK
from pathlib								import Path
from typing									import List
from typing									import Optional
from pygwarts.magical.chests				import Chest
from pygwarts.magical.spells				import patronus
from pygwarts.hagrid.cultivation.sifting	import SiftingController








class Tree(Chest):

	"""
		hagrid core class, that represents destination object. Appear to be an object that is a target for
		every "planting" (processing), that hagrid handles, so every Tree object must correspond to only
		one single folder. Must have one mandatory and three optional fields:
			bough			- destination root directory;
			twigs			- optional SiftingController object for folders filtering;
			leafs			- optional SiftingController object for files filtering;
			strict_thrive	- boolean flag that decides whether or not creating of bough is allowed in case
							bough is non existent directory. If set to True, there will be attempt to create
							bough directory, including all parents. If such attempt will fail or flag is set
							to False and bough directory does not exist, tree will be discarded (means Tree
							object, as Chest object, will invoke "unload" method to become empty Chest, so
							any Bloom object that must be assigned to it's Tree object for "planting" will
							be discarded and won't "plant" current Tree object, cause it's "boughless").
		Mandatory "bough" field might be either a string, that represents destination root folder absolute
		path, or corresponding Path object. In first case string will be converted to a Path object, anyway,
		cause all "planting" relies on every Tree object will have it's "bough" as a Path object.
		Tree class is a child of a Chest class, so actually Tree object will be a container. As Tree object
		is a target of any "planting" processes, the original design assumes Tree object will content Bloom
		objects, that organizes "planting". Bloom objects implemented the way they will reach every Tree
		object they can and put itself into it, so that Tree object will be "planted" by that Bloom objects.
		Tree object might also be encapsulated in Copse object, which is like a Chest for a Tree objects. In
		initiation time Tree object will check if it's upper layer is a Copse object to put itself in it.
		Any exception during Tree initiation will lead to discarding.
	"""

	bough			:str | Path
	twigs			:Optional[SiftingController]
	leafs			:Optional[SiftingController]
	strict_thrive	:Optional[bool]


	def twigify(self, twigs :List[Path]) -> List[Path] :

		"""
			Utility method that serves as a current Tree object personal SiftingController invoker
			for "twigs" sifting (folders filtering).
			Accepts list of Path objects, that represents folders to be sifted (filtered).
			This method will search for escalatable for current Tree object SiftingController named
			"twigs" to invoke it with provided argument "twigs" to execute sifting. That "controller"
			must return list of Path objects, that will represent sifting result, so it will be the
			value to return by "twigify". If "controller" is not found, original "twigs" list returned.
		"""

		if	callable((twigs_sift := getattr(self, "twigs", None))):
			tree_twigs = twigs_sift(twigs, thriving=True)


			if	(twigifies := len(tree_twigs) != len(twigs)):
				self.loggy.debug(f"Number of twigifies: {len(tree_twigs)}")


			return	tree_twigs
		return		twigs




	def leafify(self, leafs :List[Path]) -> List[Path] :

		"""
			Utility method that serves as a current Tree object personal SiftingController invoker
			for "leafs" sifting (files filtering).
			Accepts list of Path objects, that represents files to be sifted (filtered).
			This method will search for escalatable for current Tree object SiftingController named
			"leafs" to invoke it with provided argument "leafs" to execute sifting. That "controller"
			must return list of Path objects, that will represent sifting result, so it will be the
			value to return by "leafify". If "controller" is not found, original "leafs" list returned.
		"""

		if	callable((leafs_sift := getattr(self, "leafs", None))):
			tree_leafs = leafs_sift(leafs)


			if	(leafifies := len(tree_leafs) != len(leafs)):
				self.loggy.debug(f"Number of leafifies: {len(tree_leafs)}")


			return	tree_leafs
		return		leafs




	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		try:


			if	isinstance(self.bough, str): self.bough = Path(self.bough)

			if	not isinstance(self.bough, Path) or (
				not self.bough.is_dir() and getattr(self, "strict_thrive", True)
			):	self.loggy.info(f"Bough \"{self.bough}\" is invalid")


			else:


				# At this point "bough" is a Path object anyway, or this point will be skipped and tree
				# discarded its blooms. If "bough" was a Path object which points to not existent directory,
				# "strict_thrive" not assigned or set to True causes discard too, but at this point both
				# or any of that conditions was not met. Whether "bough" path will be created by non
				# restrictive "strict_thrive", or it is already existent directory, it must be a writable
				# directory for "planting" to happen, otherwise Exception will discard tree's blooms.
				# Discarded blooms will not "plant" current tree.
				self.bough.mkdir(parents=True, exist_ok=True)
				assert osaccess(self.bough, W_OK), f"Bough \"{self.bough}\" thrive restricted"


				if	hasattr(self, "_UPPER_LAYER"):
					if	isinstance(self._UPPER_LAYER, Copse):


						self._UPPER_LAYER(self)
						self.loggy.debug(f"Thriving tree in {self._UPPER_LAYER} Copse")
					else:
						self.loggy.debug("Not thrivable upper layer")
				else:	self.loggy.debug("Thriving single tree")


				return
		except	Exception as E:	self.loggy.error(f"Tree {self} can't thrive due to {patronus(E)}")


		self.unload()
		self.loggy.debug("Discarded")








class Copse(Chest):

	"""
		hagrid core utility class, that allows Tree (destination) objects encapsulation.
		As this object is Chest child, it is a container for Tree objects. Every Tree objects, declared
		right under Copse object, will detect it and put itself in it's upper layer Copse object.
		The Copse object itself might be encapsulated in another Copse object, so in initiation time
		current object will check if it's upper layer also a Copse object, to put itself in it.
		Doesn't have mandatory fields, but shares optional fields with Tree class:
			twigs			- optional SiftingController object for folders filtering;
			leafs			- optional SiftingController object for files filtering;
			strict_thrive	- boolean flag that decides whether or not creating of bough is allowed in case
							bough is non existent directory for any Tree object encapsulated. If set to True,
							there will be attemp to create bough directory, including all parents, for every
							Tree object. If such attempt will fail or flag is set to False and bough directory
							does not exist, tree will be discarded (means Tree object, as Chest object, will
							invoke "unload" method to become empty Chest, so any Bloom object that must be
							assigned to it's Tree object for "planting" will be discarded and won't "plant"
							current Tree object, cause it's "boughless").
		The idea behind such encapsulations has to sides:
			Association	- once optional "twigs" or/and "leafs" controllers declared for Copse object, every
						Tree object encapsulated will have access to that controllers. Also Bloom objects,
						declared for Copse objects (declaration must take place right after all Tree objects,
						cause of mutable chain order; every Tree initiation includes Copse encapsulation, so
						it's must go before any Bloom object will iterate over Copse to find Trees) will also
						affect Tree objects encapsulated. So if some Tree objects have something mutual, Copse
						is the way to go.
			Isolation	- The opposite way is to isolate some objects from another. It can be used in anyway.
						The main example for such behavior is separating *"global" and "local" sifting. First
						one implies filtering for every Tree to be "planted", while "local" is an auxiliary
						option. In their difference lies the necessity of isolated declaration - "global" must
						have it's very own "controllers", which are reachable only for *object that implements
						"global" sifting; "local" might be declared anywhere except *"global" object, with
						only restriction that it must be reachable for the Tree objects it is belong to. As
						the default behavior of *"global" object and Tree objects include the phase of
						automatic sifting invocation, Tree objects might be encapsulated to Copse object to
						provide common "controllers", and also to isolate some Tree objects form another
						and from *"global" object, that Copse might be encapsulated in any more Copse objects.
						Last idea will allow isolate "controllers" of Tree objects from *"global" object when
						the last one doesn't have it's own, and they will be search by escalation. Anyway,
						the situation when *"global" object and Tree object share "controllers" is not
						dangerous, and, when fields are well organized, might only lead to redundant sifting
						for already sifted items.

		*	- refer to hagrid.planting.Flourish
	"""

	twigs			:Optional[SiftingController]
	leafs			:Optional[SiftingController]
	strict_thrive	:Optional[bool]


	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		if	hasattr(self, "_UPPER_LAYER"):
			if	isinstance(self._UPPER_LAYER, Copse):


				self.loggy.debug(f"Soiling inner copse in {self._UPPER_LAYER} Copse")
				self._UPPER_LAYER(self)
			else:
				self.loggy.debug("Not thrivable upper layer")
		else:	self.loggy.debug("Outer Copse soiled")










from pygwarts.irma.access.inducers	import AccessInducer








def posnum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a positive floatable "value", as a second argument, and return it.
		Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if 0 <float(str(value)) else None
	except:	return








def negnum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a negative floatable "value", as a second argument, and return it.
		Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if float(str(value)) <0 else None
	except:	return








def nonnegnum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a non-negative floatable "value", as a second argument, and return it.
		Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if 0 <= float(str(value)) else None
	except:	return








def nonposnum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a non-positive floatable "value", as a second argument, and return it.
		Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if float(str(value)) <= 0 else None
	except:	return








def zeronum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a zero floatable "value", as a second argument, and return it.
		Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if float(str(value)) == 0. else None
	except:	return








def plurnum(caller :AccessInducer, value :int | float | str) -> int | float | str :

	"""
		Helper function that filters a positive floatable "value", as a second argument, that greater
		than one and return it. Returns None in case of "value" is not floatable or if it is not positive.
		By floatable value the valid result of str -> float cast is meant.
		Originally designed to accept a Transmutable object (AccessInducer) as a first argument "caller".
	"""

	try:	return value if 1 <float(str(value)) else None
	except:	return










import	re
from	pathlib								import Path
from	typing								import Dict
from	typing								import List
from	typing								import Tuple
from	typing								import Callable
from	pygwarts.magical.spells				import patronus
from	pygwarts.hagrid.bloom				import Bloom
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class Efflorescence(Bloom):

	"""
		Implementation of Bloom object that processes both folders and files, but at destination (bough)
		and by processing removing is meant. Complies the "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.

		++++++++++++++	ATTENTION
		++++++++++++++	MUST
		++++++++++++++	BE
		++++++++++++++	PAID
		++++++++++++++	FURTHER

		As current class serves as an instrument, which will cause

											UNRECOVERABLE

											REMOVING

											OF

											DATA

		please read carefully following doc string section to prevent ANY UNWANTED CONSEQUENCES!

		This class consists of __call__, which include proper "planting" organization, and if no obstacles
		are found - initiates bough twigs/leafs (target folders/files) removing separately, by invoking
		crucial method "sprigs_trimming". Along with many arguments, which "sprigs_trimming" accepts,
		there are "sprout_sprigs" and "bough_sprigs", which must be explained a lot.
		There is some difference between "regular" Bloom object, like *Rejuvenation or **Germination, which
		takes corresponding leafs or twigs of source folder and implements "planting" to target folder.
		In "Efflorescence", for any "bough sprig" (target folder/file) to be removed, it is must be checked
		against corresponding "sprout sprig" (source folder/file), which means, that if current source branch,
		that is currently walked, exists at target location and there are some "bough sprigs" that aren't
		represented as "sprout sprigs", those "bough sprigs" will be removed. In order for such "planting"
		to be possible, "sprigs_trimming" method takes "sprout_sprigs", that are "bloom-interface" arguments
		"twigs" and "leafs", and takes "bough_sprigs", that are obtained from "current_bough" branch (target
		folder, that is correspond to current source folder, by taking source folder relative to source root
		folder path and join it to target root folder) lists of Path objects as "bough_twigs" and
		"bough_leafs", to first produce, by helper method "ingather", the list of "trimmable_sprigs", which
		is a list of Path objects that represent "twigs" and "leafs" of bough, that are not represented in
		sprout, so they can be removed, and actually removes them.
		All operations, described above, are guarded by 2 mandatory conditions for "Efflorescence" to satisfy:
			branches	- escalatable field, that must be a dictionary, that maps any "bough", that must be
						effloresced, as a string object with a tuple of strings, that represents "sprouts",
						against which efflorescence is allowed. In other words, that dictionary keys must be
						a target root folder path strings, and their values must be tuples with source folders
						path strings. Any sprout-bough pair will be checked for presence in "branches", and
						efflorescence will proceed only if this condition is satisfied. This dictionary values
						must only be tuples to ensure it was obtained the right way only.
			controllers	- for every "origin" (Tree object) to be effloresced, there must be corresponding
						escalatable "SiftingController" objects (mandatory named as twigs/leafs), that must
						have "include_field" initialized and content the item for current "origin" bough
						to be matched. In other words, the Tree object will be effloresced if it's
						corresponding "twigs" and "leafs" "controllers" will exist and will have it's
						"include_field" organized to have a match for current bough branch. Also exclude_field
						might be used to exclude some bough branches from Efflorescence.
		The last thing required by "Efflorescence" is a typical for Bloom object condition, that mandatory
		callable "trim" must be escalatable for every "origin" and must implement "planting interface"
		to accept following arguments:
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			sprig	- Path object that represents current "planting" source folder/file;
			bough	- Path object that represents current "planting" destination parent folder.
		Any Exception raised by "trim" might not be handled by "trim", as originally designed, but will be
		handled by this Bloom object. In this case, after the Exception will be handled, the iteration over
		leafs list will be continued. The iteration will be stopped if either "origin" argument is not a
		Tree object, or "origin" is a Tree object but with invalid "bough" field, or the "trim"
		implementation either not callable or not found.
		 ____________________________________________________________________________________________________
		|											RESUMING												 |
		|____________________________________________________________________________________________________|
		| The "Efflorescence" object is a callable, that implements special interface, to remove target      |
		| folders and files. Dictionary "branches" must be a reachable for "Efflorescence" object field,     |
		| that content target for folders strings mapped with tuples of source folders strings. Every Tree   |
		| object, as target, must have reachable objects "twigs" and "leafs" for execution of corresponding  |
		| folders and files removing. Each Tree object's "twigs" and "leafs" must have "include_field" that  |
		| content strings or regular expressions for target location items to be removed. "exclude_field" of |
		| every Tree object might also content strings or regular expressions for target location items to   |
		| to not be removed. The removing relies on callable object "trim", that implements special interface|
		| to handle folders and files removing.                                                              |
		|____________________________________________________________________________________________________|

		*	- refer to hagrid.bloom.leafs.Rejuvenation
		**	- refer to hagrid.bloom.leafs.Germination
	"""

	branches	:Dict[str, Tuple[str,...]]

	def ingather(self, sprout_sprigs :List[Path], bough_sprigs :List[Path]) -> List[Path] :

		"""
			Helper method that provides the way to fetch the list of sprigs to be trimmed as Path objects,
			by maintaining the difference between set of sprig names at current bough branch against set
			of sprig names at current sprout branch.
			As "bloom-interface" operates with "twigs" and "leafs" lists of Path objects, and the difference
			between sprout branch and bough branch is the corresponding Path objects "name" properties, as
			corresponding folders/files relative paths, current method produces list of Path objects,
			that are correspond to difference between set of current source names and target names.
		"""

		if	isinstance(sprout_sprigs, list) and isinstance(bough_sprigs, list):

			sprout_map	= { sprig.name: sprig for sprig in sprout_sprigs }
			bough_map	= { sprig.name: sprig for sprig in bough_sprigs }
			return		[ bough_map[name] for name in set(bough_map).difference(sprout_map) ]

		self.loggy.warning(f"{self} failed to ingather sprout and bough sprigs")




	def sprigs_trimming(
							self,
							origin			:Tree,
							sprout			:str,
							branch			:Path,
							bough_sprigs	:List[Path],
							bough			:Path,
							sprout_sprigs	:List[Path],
							trim_type		:str,
							trimmer			:Callable[[Tree, str, Path, Path, Path], None],
						):

		"""
			Efflorescence core method, which in charge for actual folders and files removing.
			Accepts following positional arguments:
				origin			- current Tree object to be processed;
				sprout			- current source root folder as a string;
				branch			- current source working directory as a Path object;
				bough_sprigs	- list of Path objects that represents current target working folder
								items (folders and files);
				bough			- current target root folder as a Path object;
				sprout_sprigs	- list of Path objects that represents current source working folder
								items (folders and files);
				trim_type		- string that represents the type of items currently processing, "twigs"
								for folders and "leafs" for files;
				trimmer			- callable "planting" object, that handles removing operation.
			Validates current "origin" to have corresponding valid "controllers". Next initiates that
			"controllers" filtering to obtain list of actual items to be removed. At last iterates
			over those "trimmable_sprigs" list and invoke provided "trimmer" callable, which must
			execute actual removing and must implements the "planting interface" to accept arguments:
				origin	- Tree object that is currently in "planting";
				sprout	- string that represents the current source root;
				branch	- Path object that represents current source folder;
				leaf	- Path object that represents current operation source file;
				bough	- Path object that represents current operation destination parent folder.
			Returns None.
		"""

		if	isinstance((controller := getattr(origin, trim_type, None)), SiftingController):
			if	hasattr(controller, "include_field"):


				trimmable_sprigs = self.ingather(sprout_sprigs, controller(bough_sprigs))
				self.loggy.debug(f"Number of trimmable {trim_type}: {len(trimmable_sprigs)}")


				for sprig in trimmable_sprigs:
					try:

						trimmer(origin, sprout, branch, sprig, bough)
						self.mark_tree(str(origin), "trim")
						self.loggy.debug(f"Effloresced \"{sprig}\"")


					except	Exception as E:

						self.loggy.warning(f"Failed to effloresce \"{sprig}\" due to {patronus(E)}")
			else:		self.loggy.debug(f"Include field for {trim_type} not found")
		else:			self.loggy.debug(f"Controller for {trim_type} not found")




	def includable(self, bough :str) -> Tuple[str,...] :

		"""
			Helper method that serves as a validator for bough-sprout pairs in "branches" dictionary.
			If "bough" string, that must represent Tree object bough field, is an existent key in
			"branches", which mapped with a tuple of strings, that must represent corresponding sprouts
			mapped with current "bough" for efflorescence, that tuple will be returned.
			In any other cases logs corresponding debug message and returns empty tuple.
		"""

		if	isinstance(getattr(self, "branches", None), dict):
			if	(sprouts := self.branches.get(str(bough))) is not None:
				if	isinstance(sprouts, tuple):


					return	sprouts
				else:
					self.loggy.debug(f"Invalid sprouts mapping for \"{bough}\"")
			else:	self.loggy.debug(f"Bough \"{bough}\" not included for Efflorescence")
		else:		self.loggy.debug("No branches included for Efflorescence")
		return		tuple()




	def __call__(self, origin :Tree, sprout :str, branch :Path, twigs :List[Path], leafs :List[Path]):


		if	isinstance(getattr(self, "branches", None), dict) and isinstance(sprout, str):
			if	isinstance(origin, Tree) and isinstance((bough := getattr(origin, "bough", None)), Path):


				# IMPORTANT: "trim" to be invoked from "origin" to ensure right escalation, as "planting"
				# invocation must be issued not from Bloom object but from Tree object to be processed.
				# Further "planting" invocation relies on callable "trim", that escalated from
				# the current Tree object, to implement "planting interface". The chain of calls,
				# that will lead to that moment assumes all arguments are valid, cause the original
				# design of "planting" object doesn't involve arguments validation. That means
				# only "origin" validation at the beginning guards possible Exceptions, that might
				# be raised in "trim" because of arguments inconsistency, and which all won't be
				# handled by original "planting" design but will be caught back in current Bloom,
				# by special method "sprigs_trimming" implementation.
				if	callable(trim := getattr(origin, "trim", None)):
					if	(current_bough := bough.joinpath(branch.relative_to(sprout))).is_dir():
						self.loggy.debug(f"Commencing efflorescence of \"{current_bough}\"")


						sprigs		= list(current_bough.iterdir())
						bough_twigs	= list(filter(Path.is_dir,	sprigs))
						bough_leafs	= list(filter(Path.is_file,	sprigs))


						if	len(bough_twigs) or len(bough_leafs):
							if	sprout in self.includable(bough):


								self.sprigs_trimming(

									origin,
									sprout,
									branch,
									bough_twigs,
									bough,
									twigs,
									"twigs",
									trim
								)
								self.sprigs_trimming(

									origin,
									sprout,
									branch,
									bough_leafs,
									bough,
									leafs,
									"leafs",
									trim
								)
							else:	self.loggy.debug(f"Sprout \"{sprout}\" not included for Efflorescence")
						else:		self.loggy.debug(f"No sprigs to effloresce \"{current_bough}\"")
					else:			self.loggy.debug(f"Bough \"{current_bough}\" not thrived")
				else:				self.announce_implement(str(origin), "trim")
			else:					self.announce_bough(str(origin), "effloresce")
		else:						self.loggy.debug("Branches not found or sprout is not a string")










import	unittest
from	typing							import Dict
from	typing							import Callable
from	typing							import Generator
from	datetime						import datetime
from	datetime						import timedelta
from	pygwarts.magical.time_turner	import TimeTurner
from	pygwarts.tests.magical			import MagicalTestCase








class InitiationCases(MagicalTestCase):

	"""
		Comprehensive initiation set
	"""

	def init_datepoint_matcher(
									self,
									turner	:Callable[[int,int,int],TimeTurner],
									pattern	:Callable[[int,int,int],str],
								):

		for full_year in self.FYEARS:
			for month in self.MONTHS:
				for day	in self.DAYS_31:
					with self.subTest(d=day, m=month, Y=full_year):

						# This looks like a weak method of testing...
						# Must be some more robust idea.
						try:	self.assertEqual(turner(day,month,full_year),pattern(day,month,full_year))
						except	ValueError:
							self.assertRaisesRegex(

								ValueError,
								"day is out of range for month",
								datetime,
								full_year, month, day
							)




	def init_timepoint_matcher(
									self,
									turner	:Callable[[int,int,int],TimeTurner],
									pattern	:Callable[[int,int,int],str],
								):

		"""
			self.HOURS and self.MINSEC are simplified
		"""

		for hour in ( 0, 10, 23 ):
			for minute in ( 0, 15, 33, 59 ):
				for second in ( 0, 28, 42, 59 ):
					with self.subTest(H=hour, M=minute, S=second):
						self.assertEqual(turner(hour, minute, second), pattern(hour, minute, second))








	def test_datepoint_sjoint_1(self):

		for sep in self.DELIMETERS:
			self.init_datepoint_matcher(

				(lambda d,m,Y : TimeTurner(f"{d}{sep}{m}{sep}{Y}").dmY_aspath),
				(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
			)
	def test_datepoint_sjoint_2(self):

		for sep in self.DELIMETERS:
			self.init_datepoint_matcher(

				(lambda d,m,Y : TimeTurner(
					f"{m}{sep}{d}{sep}{Y}").dmY_aspath
					if d <13 else
					TimeTurner(f"{m}{sep}{d}{sep}{Y}").mdY_aspath
				),
				(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
			)
	def test_datepoint_sjoint_3(self):

		for sep in self.DELIMETERS:
			self.init_datepoint_matcher(

				(lambda d,m,Y : TimeTurner(f"{Y}{sep}{m}{sep}{d}").Ymd_aspath),
				(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
			)
	def test_datepoint_sjoint_4(self):

		for sep in self.DELIMETERS:
			self.init_datepoint_matcher(

				(lambda d,m,Y : TimeTurner(
					f"{Y}{sep}{d}{sep}{m}").Ymd_aspath
					if d <13 else
					TimeTurner(f"{Y}{sep}{d}{sep}{m}").Ydm_aspath
				),
				(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
			)




	def test_datepoint_iter_int_day_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ d, str(m), str(Y) ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(m), d, str(Y) ]).dmY_aspath if d <13 else TimeTurner([ str(m), d, str(Y) ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ str(Y), str(m), d ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_day_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(Y), d, str(m) ]).Ymd_aspath if d <13 else TimeTurner([ str(Y), d, str(m) ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_month_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(d), m, str(Y) ]).dmY_aspath
				),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_month_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ m, str(d), str(Y) ]).dmY_aspath if d <13 else TimeTurner([ str(m), d, str(Y) ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_month_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ str(Y), m, str(d) ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_month_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(Y), str(d), m ]).Ymd_aspath if d <13 else TimeTurner([ str(Y), str(d), m ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_fyear_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ str(d), str(m), Y ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_fyear_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(m), str(d), Y ]).dmY_aspath if d <13 else TimeTurner([ str(m), str(d), Y ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_fyear_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ Y, str(m), str(d) ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_fyear_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ Y, str(d), str(m) ]).Ymd_aspath if d <13 else TimeTurner([ Y, str(d), str(m) ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_day_month_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ d, m, str(Y) ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_month_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ m, d, str(Y) ]).dmY_aspath if d <13 else TimeTurner([ m, d, str(Y) ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_month_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ str(Y), m, d ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_day_month_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(Y), d, m ]).Ymd_aspath if d <13 else TimeTurner([ str(Y), d, m ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_day_fyear_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ d, str(m), Y ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_fyear_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ str(m), d, Y ]).dmY_aspath if d <13 else TimeTurner([ str(m), d, Y ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_fyear_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ Y, str(m), d ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_day_fyear_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ Y, d, str(m) ]).Ymd_aspath if d <13 else TimeTurner([ Y, d, str(m) ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_month_fyear_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ str(d), m, Y ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_month_fyear_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ m, str(d), Y ]).dmY_aspath if d <13 else TimeTurner([ m, str(d), Y ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_month_fyear_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ Y, m, str(d) ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_month_fyear_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ Y, str(d), m ]).Ymd_aspath if d <13 else TimeTurner([ Y, str(d), m ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_iter_int_day_month_fyear_1(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ d, m, Y ]).dmY_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(str(d).zfill(2), str(m).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_month_fyear_2(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ m, d, Y ]).dmY_aspath if d <13 else TimeTurner([ m, d, Y ]).mdY_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(str(m).zfill(2), str(d).zfill(2),Y))
		)
	def test_datepoint_iter_int_day_month_fyear_3(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner([ Y, m, d ]).Ymd_aspath),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(m).zfill(2), str(d).zfill(2)))
		)
	def test_datepoint_iter_int_day_month_fyear_4(self):
		self.init_datepoint_matcher(

			(lambda d,m,Y : TimeTurner(
				[ Y, d, m ]).Ymd_aspath if d <13 else TimeTurner([ Y, d, m ]).Ydm_aspath
			),
			(lambda d,m,Y : "%s/%s/%s"%(Y,str(d).zfill(2), str(m).zfill(2)))
		)




	def test_datepoint_epoch_init(self):

		for full_year in ( 2019,2020 ):
			for month in ( 2,6,10 ):
				for day in ( 1,7,10,22 ):
					for hour in ( 0,3,13,22 ):
						for minute in ( 0,1,11,33,42,59 ):
							for second in ( 0,2,22,36,59 ):
								with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

									stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
									self.assertIsInstance(stamp, float)
									self.assertEqual(TimeTurner(stamp).epoch, stamp)
									self.assertEqual(TimeTurner(int(stamp)).epoch, stamp)




	def test_datepoint_TimeTurner_init(self):

		full_year = 2019
		for month in ( 2,10 ):
			for day in ( 1,11,22 ):
				for hour in ( 0,13 ):
					for minute in ( 42,59 ):
						for second in ( 0,36 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(TimeTurner(TimeTurner(stamp)).epoch, stamp.timestamp())




	def test_datepoint_datetime_init(self):

		full_year = 2020
		for month in ( 6,2,5 ):
			for day in ( 7,10,25 ):
				for hour in (3,22 ):
					for minute in ( 0,59 ):
						for second in ( 36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(TimeTurner(stamp).epoch, stamp.timestamp())








	def test_timepoint_HM_sjoint(self):

		for sep in list(self.DELIMETERS) + [ "" ]:
			self.init_timepoint_matcher(

				(lambda H,M,_ : TimeTurner(
					timepoint="%s%s%s"%(str(H).zfill(2), sep, str(M).zfill(2))).HM_asjoin
				),
				(lambda H,M,_ : "%s%s"%(str(H).zfill(2), str(M).zfill(2)))
			)




	def test_timepoint_HM_iter_int_H(self):
		self.init_timepoint_matcher(

			(lambda H,M,_ : TimeTurner(timepoint=[ H,str(M) ]).HM_asjoin),
			(lambda H,M,_ : "%s%s"%(str(H).zfill(2), str(M).zfill(2)))
		)
	def test_timepoint_HM_iter_int_M(self):
		self.init_timepoint_matcher(

			(lambda H,M,_ : TimeTurner(timepoint=[ str(H),M ]).HM_asjoin),
			(lambda H,M,_ : "%s%s"%(str(H).zfill(2), str(M).zfill(2)))
		)
	def test_timepoint_HM_iter_int_HM(self):
		self.init_timepoint_matcher(

			(lambda H,M,_ : TimeTurner(timepoint=[ H,M ]).HM_asjoin),
			(lambda H,M,_ : "%s%s"%(str(H).zfill(2), str(M).zfill(2)))
		)




	def test_timepoint_HMS_sjoint(self):

		for sep in list(self.DELIMETERS) + [ "" ]:
			self.init_timepoint_matcher(

				(lambda H,M,S : TimeTurner(

					timepoint="%s%s%s%s%s"%(

						str(H).zfill(2), sep, str(M).zfill(2), sep, str(S).zfill(2))

					).HMS_asjoin
				),
				(lambda H,M,S : "%s%s%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
			)




	def test_timepoint_HMS_iter_int_H(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ H,str(M),str(S) ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_M(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ str(H),M,str(S) ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_S(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ str(H),str(M),S ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_HM(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ H,M,str(S) ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_HS(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ H,str(M),S ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_MS(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ str(H),M,S ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)
	def test_timepoint_HMS_iter_int_HMS(self):
		self.init_timepoint_matcher(

			(lambda H,M,S : TimeTurner(timepoint=[ H,M,S ]).HMS_ascolon),
			(lambda H,M,S : "%s:%s:%s"%(str(H).zfill(2), str(M).zfill(2), str(S).zfill(2)))
		)




	def test_timepoint_epoch_init(self):

		stamp = datetime.now()
		self.assertIsInstance(stamp.timestamp(), float)
		self.assertEqual(TimeTurner(timepoint=stamp.timestamp()).epoch, stamp.timestamp())
		self.assertEqual(TimeTurner(timepoint=int(stamp.timestamp())).epoch, int(stamp.timestamp()))
	def test_timepoint_TimeTurner_init(self):

		stamp = datetime.now()
		self.assertEqual(TimeTurner(TimeTurner(timepoint=stamp)).epoch, stamp.timestamp())
	def test_timepoint_datetime_init(self):

		stamp = datetime.now()
		self.assertEqual(TimeTurner(timepoint=stamp).epoch, stamp.timestamp())
















	def test_datepoint_sjoint_timepoint_sjoint_epoch_match(self):

		full_year = 1991
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(
										f"{day}/{month}/{full_year}",f"{hour}:{minute}:{second}"
									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_sjoint_timepoint_iter_epoch_match(self):

		full_year = 1992
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(
										f"{day}/{month}/{full_year}",[ hour, minute, second ]
									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_sjoint_timepoint_epoch_epoch_match(self):

		full_year = 1993
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(
									TimeTurner(f"{day}/{month}/{full_year}", stamp).epoch, stamp
								)
								self.assertEqual(

									TimeTurner(f"{day}/{month}/{full_year}", int(stamp)).epoch,
									int(stamp)
								)




	def test_datepoint_sjoint_timepoint_TimeTurner_epoch_match(self):

		full_year = 1994
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(

									TimeTurner(f"{day}/{month}/{full_year}", TimeTurner(stamp)).epoch,
									stamp
								)




	def test_datepoint_sjoint_timepoint_datetime_epoch_match(self):

		full_year = 1995
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(

									TimeTurner(f"{day}/{month}/{full_year}", stamp).epoch,
									stamp.timestamp()
								)








	def test_datepoint_iter_timepoint_sjoint_epoch_match(self):

		full_year = 1996
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(

										[ day, month, full_year ], f"{hour}:{minute}:{second}"
									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_iter_timepoint_iter_epoch_match(self):

		full_year = 1997
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(

										[ day, month, full_year ], [ hour, minute, second ]
									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_iter_timepoint_epoch_epoch_match(self):

		full_year = 1998
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(
									TimeTurner([ day, month, full_year ], stamp).epoch, stamp
								)
								self.assertEqual(
									TimeTurner([ day, month, full_year ], int(stamp)).epoch, int(stamp)
								)




	def test_datepoint_iter_timepoint_TimeTurner_epoch_match(self):

		full_year = 1999
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(

									TimeTurner([ day, month, full_year ], TimeTurner(stamp)).epoch,
									stamp
								)




	def test_datepoint_iter_timepoint_datetime_epoch_match(self):

		full_year = 2000
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(

									TimeTurner([ day, month, full_year ], stamp).epoch,
									stamp.timestamp()
								)








	def test_datepoint_epoch_timepoint_sjoint_epoch_match(self):

		full_year = 2001
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(
									TimeTurner(stamp, f"{hour}:{minute}:{second}").epoch, stamp
								)
								self.assertEqual(
									TimeTurner(int(stamp), f"{hour}:{minute}:{second}").epoch, stamp
								)




	def test_datepoint_epoch_timepoint_iter_epoch_match(self):

		full_year = 2002
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(
									TimeTurner(stamp, [ hour, minute, second ]).epoch, stamp
								)
								self.assertEqual(
									TimeTurner(int(stamp), [ hour, minute, second ]).epoch, stamp
								)




	def test_datepoint_epoch_timepoint_epoch_epoch_match(self):

		full_year = 2003
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(TimeTurner(stamp, stamp).epoch, stamp)
								self.assertEqual(TimeTurner(int(stamp), stamp).epoch, stamp)
								self.assertEqual(TimeTurner(stamp, int(stamp)).epoch, int(stamp))
								self.assertEqual(TimeTurner(int(stamp), int(stamp)).epoch, int(stamp))




	def test_datepoint_epoch_timepoint_TimeTurner_epoch_match(self):

		full_year = 2004
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(TimeTurner(stamp, TimeTurner(stamp)).epoch, stamp)
								self.assertEqual(TimeTurner(int(stamp), TimeTurner(stamp)).epoch, stamp)




	def test_datepoint_epoch_timepoint_datetime_epoch_match(self):

		full_year = 2005
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(
									TimeTurner(stamp.timestamp(), stamp).epoch, stamp.timestamp()
								)
								self.assertEqual(
									TimeTurner(int(stamp.timestamp()), stamp).epoch, stamp.timestamp()
								)








	def test_datepoint_TimeTurner_timepoint_sjoint_epoch_match(self):

		full_year = 2006
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(

										TimeTurner([ day, month, full_year ]),
										f"{hour}:{minute}:{second}"

									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_TimeTurner_timepoint_iter_epoch_match(self):

		full_year = 2007
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								self.assertEqual(

									TimeTurner(

										TimeTurner([ day, month, full_year ]),
										[ hour, minute, second ]

									).epoch,
									datetime(full_year, month, day, hour, minute, second).timestamp()
								)




	def test_datepoint_TimeTurner_timepoint_epoch_epoch_match(self):

		full_year = 2008
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(

									TimeTurner(

										TimeTurner([ day, month, full_year ]), stamp
									).epoch, stamp
								)
								self.assertEqual(

									TimeTurner(TimeTurner([ day, month, full_year ]), int(stamp)
									).epoch, int(stamp)
								)




	def test_datepoint_TimeTurner_timepoint_TimeTurner_epoch_match(self):

		full_year = 2009
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second).timestamp()
								self.assertEqual(

									TimeTurner(
										TimeTurner([ day, month, full_year ]), TimeTurner(stamp)
									).epoch, stamp
								)




	def test_datepoint_TimeTurner_timepoint_datetime_epoch_match(self):

		full_year = 2010
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(

									TimeTurner(TimeTurner([ day, month, full_year ]), stamp
									).epoch, stamp.timestamp()
								)








	def test_datepoint_datetime_timepoint_sjoint_epoch_match(self):

		full_year = 2011
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(

									TimeTurner(stamp, f"{hour}:{minute}:{second}"
									).epoch, stamp.timestamp()
								)




	def test_datepoint_datetime_timepoint_iter_epoch_match(self):

		full_year = 2012
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(
									TimeTurner(stamp, [ hour, minute, second ]).epoch, stamp.timestamp()
								)




	def test_datepoint_datetime_timepoint_epoch_epoch_match(self):

		full_year = 2013
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(
									TimeTurner(stamp, stamp.timestamp()).epoch, stamp.timestamp()
								)
								self.assertEqual(

									TimeTurner(stamp, int(stamp.timestamp())).epoch,
									int(stamp.timestamp())
								)




	def test_datepoint_datetime_timepoint_TimeTurner_epoch_match(self):

		full_year = 2014
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(
									TimeTurner(stamp, TimeTurner(stamp)).epoch, stamp.timestamp()
								)




	def test_datepoint_datetime_timepoint_datetime_epoch_match(self):

		full_year = 2015
		for month in ( 2,6,10 ):
			for day in ( 1,7,10,22 ):
				for hour in ( 0,3,13,22 ):
					for minute in ( 0,1,11,33,42,59 ):
						for second in ( 0,2,22,36,59 ):
							with self.subTest(d=day, m=month, Y=full_year, H=hour, M=minute, S=second):

								stamp = datetime(full_year, month, day, hour, minute, second)
								self.assertEqual(TimeTurner(stamp, stamp).epoch, stamp.timestamp())








	def test_datetime_travel_inits(self):

		stamp = datetime(2020, 10, 20, 4, 26, 9)
		travels_keys = list(self.travels)
		L = len(travels_keys)


		def casegen(i :int, buffer :Dict[str,int]) -> Generator[Dict[str,int], None,None] :

			if	buffer : yield buffer
			for j in range(i,L):

				current_key = travels_keys[j]
				buffer[current_key] = self.travels[current_key]
				yield from casegen(j +1, buffer)

				buffer[current_key] = -self.travels[current_key]
				yield from casegen(j +1, buffer)

				del buffer[current_key]


		for case in casegen(0,{}):
			self.assertEqual(

				TimeTurner(stamp, **case).epoch,
				(stamp + timedelta(**case)).timestamp()
			)




	def test_datetime_months_travel_inits(self):

		stamp = datetime(2020, 10, 20, 4, 26, 9)

		self.assertEqual(TimeTurner(stamp, months=1).POINT, datetime(2020, 11, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=2).POINT, datetime(2020, 12, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=3).POINT, datetime(2021, 1, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=4).POINT, datetime(2021, 2, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=5).POINT, datetime(2021, 3, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=6).POINT, datetime(2021, 4, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=7).POINT, datetime(2021, 5, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=8).POINT, datetime(2021, 6, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=9).POINT, datetime(2021, 7, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=10).POINT, datetime(2021, 8, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=11).POINT, datetime(2021, 9, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=12).POINT, datetime(2021, 10, 1, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=13).POINT, datetime(2021, 11, 1, 4, 26, 9))

		self.assertEqual(TimeTurner(stamp, months=-1).POINT, datetime(2020, 9, 30, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-2).POINT, datetime(2020, 8, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-3).POINT, datetime(2020, 7, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-4).POINT, datetime(2020, 6, 30, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-5).POINT, datetime(2020, 5, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-6).POINT, datetime(2020, 4, 30, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-7).POINT, datetime(2020, 3, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-8).POINT, datetime(2020, 2, 29, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-9).POINT, datetime(2020, 1, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-10).POINT, datetime(2019, 12, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-11).POINT, datetime(2019, 11, 30, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-12).POINT, datetime(2019, 10, 31, 4, 26, 9))
		self.assertEqual(TimeTurner(stamp, months=-13).POINT, datetime(2019, 9, 30, 4, 26, 9))








	def test_datepoint_sjoint_epoch_match(self):

		for full_year in self.FYEARS:
			for month in self.MONTHS:
				for day in self.DAYS_31:
					with self.subTest(d=day, m=month, Y=full_year):

						try:

							self.assertEqual(

								TimeTurner(f"{day}/{month}/{full_year}").epoch,
								datetime(full_year, month, day).timestamp()
							)
						except	ValueError:
							self.assertRaisesRegex(

								ValueError,
								"day is out of range for month",
								datetime,
								full_year, month, day
							)




	def test_datepoint_iter_epoch_match(self):

		for full_year in self.FYEARS:
			for month in self.MONTHS:
				for day in self.DAYS_31:
					with self.subTest(d=day, m=month, Y=full_year):

						try:

							self.assertEqual(

								TimeTurner([ day, month, full_year ]).epoch,
								datetime(full_year, month, day).timestamp()
							)
						except	ValueError:
							self.assertRaisesRegex(

								ValueError,
								"day is out of range for month",
								datetime,
								full_year, month, day
							)
















class OperationsCases(MagicalTestCase):

	"""
		Operational methods and properties
	"""

	def setUp(self): self.stamp = TimeTurner([ 20, 10, 2010 ],[ 3,2,8 ])
	def test_sight(self):

		probe = self.stamp
		self.assertEqual(id(self.stamp), id(probe))

		sight = self.stamp.sight()
		self.assertEqual(self.stamp, sight)
		self.assertNotEqual(id(self.stamp), id(sight))

		sight = self.stamp.sight(months=1)
		self.assertEqual(sight.dmY_aspath, "01/11/2010")
		self.assertNotEqual(self.stamp, sight)
		self.assertNotEqual(id(self.stamp), id(sight))

		sight = self.stamp.sight(months=-1)
		self.assertEqual(sight.dmY_aspath, "30/09/2010")
		self.assertNotEqual(self.stamp, sight)
		self.assertNotEqual(id(self.stamp), id(sight))




	def test_positive_diff(self):

		probe = self.stamp
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(TimeTurner(self.stamp, minutes=-1))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, 60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(datetime(2010, 10, 20, 3, 1, 8))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, 60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(datetime(2010, 10, 20, 3, 1, 8).timestamp())
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, 60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(int(datetime(2010, 10, 20, 3, 1, 8).timestamp()))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, 60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(f"{datetime(2010, 10, 20, 3, 1, 8).timestamp()}")
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, 60.)
		self.assertEqual(id(self.stamp), id(probe))




	def test_negative_diff(self):

		probe = self.stamp
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(TimeTurner(self.stamp, minutes=1))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, -60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(datetime(2010, 10, 20, 3, 3, 8))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, -60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(datetime(2010, 10, 20, 3, 3, 8).timestamp())
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, -60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(int(datetime(2010, 10, 20, 3, 3, 8).timestamp()))
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, -60.)
		self.assertEqual(id(self.stamp), id(probe))

		diff = self.stamp.diff(f"{datetime(2010, 10, 20, 3, 3, 8).timestamp()}")
		self.assertIsInstance(diff, float)
		self.assertEqual(diff, -60.)
		self.assertEqual(id(self.stamp), id(probe))




	def test_getattr(self):

		self.assertEqual(f"{self.stamp.dmY_aspath} {self.stamp.HMS_aspath}", "20/10/2010 03/02/08")
		self.assertEqual(f"{self.stamp.dmY_aswpath} {self.stamp.HMS_aswpath}", "20\\10\\2010 03\\02\\08")
		self.assertEqual(f"{self.stamp.dmY_ascolon} {self.stamp.HMS_ascolon}", "20:10:2010 03:02:08")
		self.assertEqual(f"{self.stamp.dmY_asjoin} {self.stamp.HMS_asjoin}", "20102010 030208")
		self.assertEqual(f"{self.stamp.dmY_spaced} {self.stamp.HMS_spaced}", "20 10 2010 03 02 08")
		self.assertEqual(f"{self.stamp.dmY_dashed} {self.stamp.HMS_dashed}", "20-10-2010 03-02-08")




	def test_travels(self):

		probe = TimeTurner(self.stamp)
		self.assertEqual(self.stamp, probe)
		self.assertNotEqual(id(self.stamp), id(probe))

		traveled = probe.travel(**self.travels)
		self.assertNotEqual(self.stamp, probe)
		self.assertNotEqual(self.stamp, traveled)
		self.assertEqual(probe, traveled)
		self.assertEqual(id(probe), id(traveled))
		self.assertNotEqual(id(self.stamp), id(probe))
		self.assertNotEqual(id(self.stamp), id(traveled))




	def test_is_first_day(self):

		self.assertFalse(self.stamp.is_first_day)
		self.assertTrue(self.stamp.sight(days=-19).is_first_day)

	def test_is_leap_year(self):

		self.assertFalse(self.stamp.is_leap_year)
		self.assertTrue(self.stamp.sight(months=-24).is_leap_year)




	def test_add(self):

		probe = self.stamp + 100500
		self.assertIsInstance(probe, TimeTurner)
		self.assertNotEqual(self.stamp, probe)
		self.assertNotEqual(id(self.stamp), id(probe))
		self.assertEqual(probe.epoch, self.stamp.epoch + 100500)




	def test_sub(self):

		probe = self.stamp - 100500
		self.assertIsInstance(probe, TimeTurner)
		self.assertNotEqual(self.stamp, probe)
		self.assertNotEqual(id(self.stamp), id(probe))
		self.assertEqual(probe.epoch, self.stamp.epoch - 100500)




	def test_eq(self):

		self.assertEqual(self.stamp, TimeTurner(self.stamp))
		self.assertEqual(self.stamp, datetime(2010, 10, 20, 3, 2, 8))
		self.assertEqual(self.stamp, self.stamp.epoch)
		self.assertEqual(self.stamp, int(self.stamp.epoch))
		self.assertEqual(self.stamp, f"{self.stamp.epoch}")
		self.assertEqual(self.stamp, f"{int(self.stamp.epoch)}")




	def test_neq(self):

		self.assertNotEqual(self.stamp, TimeTurner(self.stamp, seconds=1))
		self.assertNotEqual(self.stamp, datetime(2010, 10, 20, 3, 2, 9))
		self.assertNotEqual(self.stamp, self.stamp.epoch +1)
		self.assertNotEqual(self.stamp, int(self.stamp.epoch) +1)
		self.assertNotEqual(self.stamp, f"{self.stamp.epoch +1}")
		self.assertNotEqual(self.stamp, f"{int(self.stamp.epoch) +1}")




	def test_gt(self):

		self.assertGreater(self.stamp, TimeTurner(self.stamp, seconds=-1))
		self.assertGreater(self.stamp, datetime(2010, 10, 20, 3, 2, 7))
		self.assertGreater(self.stamp, self.stamp.epoch -1)
		self.assertGreater(self.stamp, int(self.stamp.epoch) -1)
		self.assertGreater(self.stamp, f"{self.stamp.epoch -1}")
		self.assertGreater(self.stamp, f"{int(self.stamp.epoch) -1}")




	def test_ge(self):

		self.assertGreaterEqual(self.stamp, TimeTurner(self.stamp))
		self.assertGreaterEqual(self.stamp, datetime(2010, 10, 20, 3, 2, 8))
		self.assertGreaterEqual(self.stamp, self.stamp.epoch)
		self.assertGreaterEqual(self.stamp, int(self.stamp.epoch))
		self.assertGreaterEqual(self.stamp, f"{self.stamp.epoch}")
		self.assertGreaterEqual(self.stamp, f"{int(self.stamp.epoch)}")




	def test_lt(self):

		self.assertLess(self.stamp, TimeTurner(self.stamp, seconds=1))
		self.assertLess(self.stamp, datetime(2010, 10, 20, 3, 2, 9))
		self.assertLess(self.stamp, self.stamp.epoch +1)
		self.assertLess(self.stamp, int(self.stamp.epoch) +1)
		self.assertLess(self.stamp, f"{self.stamp.epoch +1}")
		self.assertLess(self.stamp, f"{int(self.stamp.epoch) +1}")




	def test_le(self):

		self.assertLessEqual(self.stamp, TimeTurner(self.stamp))
		self.assertLessEqual(self.stamp, datetime(2010, 10, 20, 3, 2, 8))
		self.assertLessEqual(self.stamp, self.stamp.epoch)
		self.assertLessEqual(self.stamp, int(self.stamp.epoch))
		self.assertLessEqual(self.stamp, f"{self.stamp.epoch}")
		self.assertLessEqual(self.stamp, f"{int(self.stamp.epoch)}")








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.irma					import IrmaTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access.utils			import byte_size_string
from	pygwarts.irma.access.utils			import WriteWrapper








class AccessUtilsCase(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_UTILS): os.remove(cls.ACCESS_UTILS)

		if	os.path.isfile(cls.wwpath): os.remove(cls.wwpath)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.ACCESS_UTILS)
		cls.wwpath = str(cls.IRMA_ROOT /"ww.txt")


	def test_number_byte_size_string(self):

		self.assertEqual(byte_size_string(0), "0B")
		self.assertEqual(byte_size_string(1), "1B")
		self.assertEqual(byte_size_string(9), "9B")
		self.assertEqual(byte_size_string(10), "10B")
		self.assertEqual(byte_size_string(99), "99B")
		self.assertEqual(byte_size_string(100), "100B")
		self.assertEqual(byte_size_string(999), "999B")
		self.assertEqual(byte_size_string(1000), "1K")
		self.assertEqual(byte_size_string(1001), "1K 1B")
		self.assertEqual(byte_size_string(1009), "1K 9B")
		self.assertEqual(byte_size_string(1099), "1K 99B")
		self.assertEqual(byte_size_string(1100), "1K 100B")
		self.assertEqual(byte_size_string(1999), "1K 999B")
		self.assertEqual(byte_size_string(99999), "99K 999B")
		self.assertEqual(byte_size_string(999999), "999K 999B")
		self.assertEqual(byte_size_string(9999999), "9M 999K 999B")
		self.assertEqual(byte_size_string(99999999), "99M 999K 999B")
		self.assertEqual(byte_size_string(999999999), "999M 999K 999B")
		self.assertEqual(byte_size_string(9999999999), "9G 999M 999K 999B")
		self.assertEqual(byte_size_string(99999999999), "99G 999M 999K 999B")
		self.assertEqual(byte_size_string(999999999999), "999G 999M 999K 999B")
		self.assertEqual(byte_size_string(9999999999999), "9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(99999999999999), "99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(999999999999999), "999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string(False), "0B")
		self.assertEqual(byte_size_string(True), "1B")

		self.assertEqual(byte_size_string(0.9), "0B")
		self.assertEqual(byte_size_string(1.9), "1B")
		self.assertEqual(byte_size_string(9.9), "9B")
		self.assertEqual(byte_size_string(10.9), "10B")
		self.assertEqual(byte_size_string(99.9), "99B")
		self.assertEqual(byte_size_string(100.9), "100B")
		self.assertEqual(byte_size_string(999.9), "999B")
		self.assertEqual(byte_size_string(1000.9), "1K")
		self.assertEqual(byte_size_string(1001.9), "1K 1B")
		self.assertEqual(byte_size_string(1009.9), "1K 9B")
		self.assertEqual(byte_size_string(1099.9), "1K 99B")
		self.assertEqual(byte_size_string(1100.9), "1K 100B")
		self.assertEqual(byte_size_string(1999.9), "1K 999B")
		self.assertEqual(byte_size_string(99999.9), "99K 999B")
		self.assertEqual(byte_size_string(999999.9), "999K 999B")
		self.assertEqual(byte_size_string(9999999.9), "9M 999K 999B")
		self.assertEqual(byte_size_string(99999999.9), "99M 999K 999B")
		self.assertEqual(byte_size_string(999999999.9), "999M 999K 999B")
		self.assertEqual(byte_size_string(9999999999.9), "9G 999M 999K 999B")
		self.assertEqual(byte_size_string(99999999999.9), "99G 999M 999K 999B")
		self.assertEqual(byte_size_string(999999999999.9), "999G 999M 999K 999B")
		self.assertEqual(byte_size_string(9999999999999.9), "9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(99999999999999.9), "99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(999999999999999.9), "999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string(.0), "0B")
		self.assertEqual(byte_size_string(.1), "0B")
		self.assertEqual(byte_size_string(.9), "0B")
		self.assertEqual(byte_size_string(.10), "0B")
		self.assertEqual(byte_size_string(.99), "0B")
		self.assertEqual(byte_size_string(.100), "0B")
		self.assertEqual(byte_size_string(.999), "0B")
		self.assertEqual(byte_size_string(.1000), "0B")
		self.assertEqual(byte_size_string(.1001), "0B")
		self.assertEqual(byte_size_string(.1009), "0B")
		self.assertEqual(byte_size_string(.1099), "0B")
		self.assertEqual(byte_size_string(.1100), "0B")
		self.assertEqual(byte_size_string(.1999), "0B")
		self.assertEqual(byte_size_string(.99999), "0B")
		self.assertEqual(byte_size_string(.999999), "0B")
		self.assertEqual(byte_size_string(.9999999), "0B")
		self.assertEqual(byte_size_string(.99999999), "0B")
		self.assertEqual(byte_size_string(.999999999), "0B")
		self.assertEqual(byte_size_string(.9999999999), "0B")
		self.assertEqual(byte_size_string(.99999999999), "0B")
		self.assertEqual(byte_size_string(.999999999999), "0B")
		self.assertEqual(byte_size_string(.9999999999999), "0B")
		self.assertEqual(byte_size_string(.99999999999999), "0B")
		self.assertEqual(byte_size_string(.999999999999999), "0B")

	def test_string_byte_size_string(self):

		self.assertEqual(byte_size_string("0"), "0B")
		self.assertEqual(byte_size_string("1"), "1B")
		self.assertEqual(byte_size_string("9"), "9B")
		self.assertEqual(byte_size_string("10"), "10B")
		self.assertEqual(byte_size_string("99"), "99B")
		self.assertEqual(byte_size_string("100"), "100B")
		self.assertEqual(byte_size_string("999"), "999B")
		self.assertEqual(byte_size_string("1000"), "1K")
		self.assertEqual(byte_size_string("1001"), "1K 1B")
		self.assertEqual(byte_size_string("1009"), "1K 9B")
		self.assertEqual(byte_size_string("1099"), "1K 99B")
		self.assertEqual(byte_size_string("1100"), "1K 100B")
		self.assertEqual(byte_size_string("1999"), "1K 999B")
		self.assertEqual(byte_size_string("99999"), "99K 999B")
		self.assertEqual(byte_size_string("999999"), "999K 999B")
		self.assertEqual(byte_size_string("9999999"), "9M 999K 999B")
		self.assertEqual(byte_size_string("99999999"), "99M 999K 999B")
		self.assertEqual(byte_size_string("999999999"), "999M 999K 999B")
		self.assertEqual(byte_size_string("9999999999"), "9G 999M 999K 999B")
		self.assertEqual(byte_size_string("99999999999"), "99G 999M 999K 999B")
		self.assertEqual(byte_size_string("999999999999"), "999G 999M 999K 999B")
		self.assertEqual(byte_size_string("9999999999999"), "9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("99999999999999"), "99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("0999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("000999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("0000999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("000000999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("0000000999999999999999"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000000999999999999999"), "999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string("0.9"), "0B")
		self.assertEqual(byte_size_string("1.9"), "1B")
		self.assertEqual(byte_size_string("9.9"), "9B")
		self.assertEqual(byte_size_string("10.9"), "10B")
		self.assertEqual(byte_size_string("99.9"), "99B")
		self.assertEqual(byte_size_string("100.9"), "100B")
		self.assertEqual(byte_size_string("999.9"), "999B")
		self.assertEqual(byte_size_string("1000.9"), "1K")
		self.assertEqual(byte_size_string("1001.9"), "1K 1B")
		self.assertEqual(byte_size_string("1009.9"), "1K 9B")
		self.assertEqual(byte_size_string("1099.9"), "1K 99B")
		self.assertEqual(byte_size_string("1100.9"), "1K 100B")
		self.assertEqual(byte_size_string("1999.9"), "1K 999B")
		self.assertEqual(byte_size_string("99999.9"), "99K 999B")
		self.assertEqual(byte_size_string("999999.9"), "999K 999B")
		self.assertEqual(byte_size_string("9999999.9"), "9M 999K 999B")
		self.assertEqual(byte_size_string("99999999.9"), "99M 999K 999B")
		self.assertEqual(byte_size_string("999999999.9"), "999M 999K 999B")
		self.assertEqual(byte_size_string("9999999999.9"), "9G 999M 999K 999B")
		self.assertEqual(byte_size_string("99999999999.9"), "99G 999M 999K 999B")
		self.assertEqual(byte_size_string("999999999999.9"), "999G 999M 999K 999B")
		self.assertEqual(byte_size_string("9999999999999.9"), "9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("99999999999999.9"), "99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("0999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("000999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("0000999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.9"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.90"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.900"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.9000"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.90000"), "999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("00000999999999999999.900000"), "999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string(".0"), "0B")
		self.assertEqual(byte_size_string(".1"), "0B")
		self.assertEqual(byte_size_string(".9"), "0B")
		self.assertEqual(byte_size_string(".10"), "0B")
		self.assertEqual(byte_size_string(".99"), "0B")
		self.assertEqual(byte_size_string(".100"), "0B")
		self.assertEqual(byte_size_string(".999"), "0B")
		self.assertEqual(byte_size_string(".1000"), "0B")
		self.assertEqual(byte_size_string(".1001"), "0B")
		self.assertEqual(byte_size_string(".1009"), "0B")
		self.assertEqual(byte_size_string(".1099"), "0B")
		self.assertEqual(byte_size_string(".1100"), "0B")
		self.assertEqual(byte_size_string(".1999"), "0B")
		self.assertEqual(byte_size_string(".99999"), "0B")
		self.assertEqual(byte_size_string(".999999"), "0B")
		self.assertEqual(byte_size_string(".9999999"), "0B")
		self.assertEqual(byte_size_string(".99999999"), "0B")
		self.assertEqual(byte_size_string(".999999999"), "0B")
		self.assertEqual(byte_size_string(".9999999999"), "0B")
		self.assertEqual(byte_size_string(".99999999999"), "0B")
		self.assertEqual(byte_size_string(".999999999999"), "0B")
		self.assertEqual(byte_size_string(".9999999999999"), "0B")
		self.assertEqual(byte_size_string(".99999999999999"), "0B")
		self.assertEqual(byte_size_string(".999999999999999"), "0B")
		self.assertEqual(byte_size_string("0.999999999999999"), "0B")
		self.assertEqual(byte_size_string("00.999999999999999"), "0B")
		self.assertEqual(byte_size_string("000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("0000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("00000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("000000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("0000000.999999999999999"), "0B")

	def test_negative_number_byte_size_string(self):

		self.assertEqual(byte_size_string(-0), "0B")
		self.assertEqual(byte_size_string(-1), "-1B")
		self.assertEqual(byte_size_string(-9), "-9B")
		self.assertEqual(byte_size_string(-10), "-10B")
		self.assertEqual(byte_size_string(-99), "-99B")
		self.assertEqual(byte_size_string(-100), "-100B")
		self.assertEqual(byte_size_string(-999), "-999B")
		self.assertEqual(byte_size_string(-1000), "-1K")
		self.assertEqual(byte_size_string(-1001), "-1K 1B")
		self.assertEqual(byte_size_string(-1009), "-1K 9B")
		self.assertEqual(byte_size_string(-1099), "-1K 99B")
		self.assertEqual(byte_size_string(-1100), "-1K 100B")
		self.assertEqual(byte_size_string(-1999), "-1K 999B")
		self.assertEqual(byte_size_string(-99999), "-99K 999B")
		self.assertEqual(byte_size_string(-999999), "-999K 999B")
		self.assertEqual(byte_size_string(-9999999), "-9M 999K 999B")
		self.assertEqual(byte_size_string(-99999999), "-99M 999K 999B")
		self.assertEqual(byte_size_string(-999999999), "-999M 999K 999B")
		self.assertEqual(byte_size_string(-9999999999), "-9G 999M 999K 999B")
		self.assertEqual(byte_size_string(-99999999999), "-99G 999M 999K 999B")
		self.assertEqual(byte_size_string(-999999999999), "-999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-9999999999999), "-9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-99999999999999), "-99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-999999999999999), "-999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string(-0.9), "0B")
		self.assertEqual(byte_size_string(-1.9), "-1B")
		self.assertEqual(byte_size_string(-9.9), "-9B")
		self.assertEqual(byte_size_string(-10.9), "-10B")
		self.assertEqual(byte_size_string(-99.9), "-99B")
		self.assertEqual(byte_size_string(-100.9), "-100B")
		self.assertEqual(byte_size_string(-999.9), "-999B")
		self.assertEqual(byte_size_string(-1000.9), "-1K")
		self.assertEqual(byte_size_string(-1001.9), "-1K 1B")
		self.assertEqual(byte_size_string(-1009.9), "-1K 9B")
		self.assertEqual(byte_size_string(-1099.9), "-1K 99B")
		self.assertEqual(byte_size_string(-1100.9), "-1K 100B")
		self.assertEqual(byte_size_string(-1999.9), "-1K 999B")
		self.assertEqual(byte_size_string(-99999.9), "-99K 999B")
		self.assertEqual(byte_size_string(-999999.9), "-999K 999B")
		self.assertEqual(byte_size_string(-9999999.9), "-9M 999K 999B")
		self.assertEqual(byte_size_string(-99999999.9), "-99M 999K 999B")
		self.assertEqual(byte_size_string(-999999999.9), "-999M 999K 999B")
		self.assertEqual(byte_size_string(-9999999999.9), "-9G 999M 999K 999B")
		self.assertEqual(byte_size_string(-99999999999.9), "-99G 999M 999K 999B")
		self.assertEqual(byte_size_string(-999999999999.9), "-999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-9999999999999.9), "-9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-99999999999999.9), "-99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string(-999999999999999.9), "-999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string(-.0), "0B")
		self.assertEqual(byte_size_string(-.1), "0B")
		self.assertEqual(byte_size_string(-.9), "0B")
		self.assertEqual(byte_size_string(-.10), "0B")
		self.assertEqual(byte_size_string(-.99), "0B")
		self.assertEqual(byte_size_string(-.100), "0B")
		self.assertEqual(byte_size_string(-.999), "0B")
		self.assertEqual(byte_size_string(-.1000), "0B")
		self.assertEqual(byte_size_string(-.1001), "0B")
		self.assertEqual(byte_size_string(-.1009), "0B")
		self.assertEqual(byte_size_string(-.1099), "0B")
		self.assertEqual(byte_size_string(-.1100), "0B")
		self.assertEqual(byte_size_string(-.1999), "0B")
		self.assertEqual(byte_size_string(-.99999), "0B")
		self.assertEqual(byte_size_string(-.999999), "0B")
		self.assertEqual(byte_size_string(-.9999999), "0B")
		self.assertEqual(byte_size_string(-.99999999), "0B")
		self.assertEqual(byte_size_string(-.999999999), "0B")
		self.assertEqual(byte_size_string(-.9999999999), "0B")
		self.assertEqual(byte_size_string(-.99999999999), "0B")
		self.assertEqual(byte_size_string(-.999999999999), "0B")
		self.assertEqual(byte_size_string(-.9999999999999), "0B")
		self.assertEqual(byte_size_string(-.99999999999999), "0B")
		self.assertEqual(byte_size_string(-.999999999999999), "0B")

	def test_negative_string_byte_size_string(self):

		self.assertEqual(byte_size_string("-0"), "0B")
		self.assertEqual(byte_size_string("-1"), "-1B")
		self.assertEqual(byte_size_string("-9"), "-9B")
		self.assertEqual(byte_size_string("-10"), "-10B")
		self.assertEqual(byte_size_string("-99"), "-99B")
		self.assertEqual(byte_size_string("-100"), "-100B")
		self.assertEqual(byte_size_string("-999"), "-999B")
		self.assertEqual(byte_size_string("-1000"), "-1K")
		self.assertEqual(byte_size_string("-1001"), "-1K 1B")
		self.assertEqual(byte_size_string("-1009"), "-1K 9B")
		self.assertEqual(byte_size_string("-1099"), "-1K 99B")
		self.assertEqual(byte_size_string("-1100"), "-1K 100B")
		self.assertEqual(byte_size_string("-1999"), "-1K 999B")
		self.assertEqual(byte_size_string("-99999"), "-99K 999B")
		self.assertEqual(byte_size_string("-999999"), "-999K 999B")
		self.assertEqual(byte_size_string("-9999999"), "-9M 999K 999B")
		self.assertEqual(byte_size_string("-99999999"), "-99M 999K 999B")
		self.assertEqual(byte_size_string("-999999999"), "-999M 999K 999B")
		self.assertEqual(byte_size_string("-9999999999"), "-9G 999M 999K 999B")
		self.assertEqual(byte_size_string("-99999999999"), "-99G 999M 999K 999B")
		self.assertEqual(byte_size_string("-999999999999"), "-999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-9999999999999"), "-9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-99999999999999"), "-99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-0999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-000999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-0000999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-000000999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-0000000999999999999999"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000000999999999999999"), "-999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string("-0.9"), "0B")
		self.assertEqual(byte_size_string("-1.9"), "-1B")
		self.assertEqual(byte_size_string("-9.9"), "-9B")
		self.assertEqual(byte_size_string("-10.9"), "-10B")
		self.assertEqual(byte_size_string("-99.9"), "-99B")
		self.assertEqual(byte_size_string("-100.9"), "-100B")
		self.assertEqual(byte_size_string("-999.9"), "-999B")
		self.assertEqual(byte_size_string("-1000.9"), "-1K")
		self.assertEqual(byte_size_string("-1001.9"), "-1K 1B")
		self.assertEqual(byte_size_string("-1009.9"), "-1K 9B")
		self.assertEqual(byte_size_string("-1099.9"), "-1K 99B")
		self.assertEqual(byte_size_string("-1100.9"), "-1K 100B")
		self.assertEqual(byte_size_string("-1999.9"), "-1K 999B")
		self.assertEqual(byte_size_string("-99999.9"), "-99K 999B")
		self.assertEqual(byte_size_string("-999999.9"), "-999K 999B")
		self.assertEqual(byte_size_string("-9999999.9"), "-9M 999K 999B")
		self.assertEqual(byte_size_string("-99999999.9"), "-99M 999K 999B")
		self.assertEqual(byte_size_string("-999999999.9"), "-999M 999K 999B")
		self.assertEqual(byte_size_string("-9999999999.9"), "-9G 999M 999K 999B")
		self.assertEqual(byte_size_string("-99999999999.9"), "-99G 999M 999K 999B")
		self.assertEqual(byte_size_string("-999999999999.9"), "-999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-9999999999999.9"), "-9T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-99999999999999.9"), "-99T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-0999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-000999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-0000999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.9"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.90"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.900"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.9000"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.90000"), "-999T 999G 999M 999K 999B")
		self.assertEqual(byte_size_string("-00000999999999999999.900000"), "-999T 999G 999M 999K 999B")

		self.assertEqual(byte_size_string("-.0"), "0B")
		self.assertEqual(byte_size_string("-.1"), "0B")
		self.assertEqual(byte_size_string("-.9"), "0B")
		self.assertEqual(byte_size_string("-.10"), "0B")
		self.assertEqual(byte_size_string("-.99"), "0B")
		self.assertEqual(byte_size_string("-.100"), "0B")
		self.assertEqual(byte_size_string("-.999"), "0B")
		self.assertEqual(byte_size_string("-.1000"), "0B")
		self.assertEqual(byte_size_string("-.1001"), "0B")
		self.assertEqual(byte_size_string("-.1009"), "0B")
		self.assertEqual(byte_size_string("-.1099"), "0B")
		self.assertEqual(byte_size_string("-.1100"), "0B")
		self.assertEqual(byte_size_string("-.1999"), "0B")
		self.assertEqual(byte_size_string("-.99999"), "0B")
		self.assertEqual(byte_size_string("-.999999"), "0B")
		self.assertEqual(byte_size_string("-.9999999"), "0B")
		self.assertEqual(byte_size_string("-.99999999"), "0B")
		self.assertEqual(byte_size_string("-.999999999"), "0B")
		self.assertEqual(byte_size_string("-.9999999999"), "0B")
		self.assertEqual(byte_size_string("-.99999999999"), "0B")
		self.assertEqual(byte_size_string("-.999999999999"), "0B")
		self.assertEqual(byte_size_string("-.9999999999999"), "0B")
		self.assertEqual(byte_size_string("-.99999999999999"), "0B")
		self.assertEqual(byte_size_string("-.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-0.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-00.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-0000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-00000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-000000.999999999999999"), "0B")
		self.assertEqual(byte_size_string("-0000000.999999999999999"), "0B")

	def test_overflow_byte_size_string(self):

		self.assertIsNone(byte_size_string(9999999999999999))
		self.assertIsNone(byte_size_string(99999999999999999))
		self.assertIsNone(byte_size_string(999999999999999999))
		self.assertIsNone(byte_size_string(9999999999999999999))
		self.assertIsNone(byte_size_string(99999999999999999999))
		self.assertIsNone(byte_size_string(999999999999999999999))
		self.assertIsNone(byte_size_string(9999999999999999999999))

		self.assertIsNone(byte_size_string(9999999999999999.9))
		self.assertIsNone(byte_size_string(99999999999999999.9))
		self.assertIsNone(byte_size_string(999999999999999999.9))
		self.assertIsNone(byte_size_string(9999999999999999999.9))
		self.assertIsNone(byte_size_string(99999999999999999999.9))
		self.assertIsNone(byte_size_string(999999999999999999999.9))
		self.assertIsNone(byte_size_string(9999999999999999999999.9))

		self.assertIsNone(byte_size_string("9999999999999999"))
		self.assertIsNone(byte_size_string("99999999999999999"))
		self.assertIsNone(byte_size_string("999999999999999999"))
		self.assertIsNone(byte_size_string("9999999999999999999"))
		self.assertIsNone(byte_size_string("99999999999999999999"))
		self.assertIsNone(byte_size_string("999999999999999999999"))
		self.assertIsNone(byte_size_string("9999999999999999999999"))
		self.assertIsNone(byte_size_string("09999999999999999999999"))
		self.assertIsNone(byte_size_string("009999999999999999999999"))
		self.assertIsNone(byte_size_string("0009999999999999999999999"))
		self.assertIsNone(byte_size_string("00009999999999999999999999"))
		self.assertIsNone(byte_size_string("000009999999999999999999999"))
		self.assertIsNone(byte_size_string("0000009999999999999999999999"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999"))
		self.assertIsNone(byte_size_string("000000009999999999999999999999"))

		self.assertIsNone(byte_size_string("9999999999999999.9"))
		self.assertIsNone(byte_size_string("99999999999999999.9"))
		self.assertIsNone(byte_size_string("999999999999999999.9"))
		self.assertIsNone(byte_size_string("9999999999999999999.9"))
		self.assertIsNone(byte_size_string("99999999999999999999.9"))
		self.assertIsNone(byte_size_string("999999999999999999999.9"))
		self.assertIsNone(byte_size_string("9999999999999999999999.9"))
		self.assertIsNone(byte_size_string("09999999999999999999999.9"))
		self.assertIsNone(byte_size_string("009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("0009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("00009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("000009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("0000009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.9"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.90"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.900"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.9000"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.90000"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.900000"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.9000000"))
		self.assertIsNone(byte_size_string("00000009999999999999999999999.90000000"))

	def test_invalid_byte_size_string(self):

		for value in (

			"forty two",[ 42 ],( 42, ),{ 42 },{ "value": 42 },
			None, ..., print, unittest, self, type, object, super(), Exception
		):
			self.assertIsNone(byte_size_string(value))








	def test_WriteWrapper(self):

		if	os.path.isfile(self.wwpath): os.remove(self.wwpath)

		@WriteWrapper(self.wwpath)
		class SomeText(Transmutable):

			def __call__(self): return "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG"
			class loggy(LibraryContrib):

				handler		= self.ACCESS_UTILS
				init_name	= "WriteWrapper"
				init_level	= 10

		self.test_case = SomeText()

		with self.assertLogs("WriteWrapper", 10) as case_loggy:
			self.assertEqual(self.test_case(), "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")

		self.assertTrue(os.path.isfile(self.wwpath))
		with open(self.wwpath) as ww:
			self.assertEqual(ww.read(), "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")

		self.assertIn("DEBUG:WriteWrapper:Writing 49 symbols", case_loggy.output)
		self.assertIn(f"INFO:WriteWrapper:Written to {self.wwpath}", case_loggy.output)




	def test_WriteWrapper_rewrite(self):

		self.fmake(self.wwpath, "I SAY\n")
		self.assertTrue(os.path.isfile(self.wwpath))
		with open(self.wwpath) as ww : self.assertEqual(ww.read(), "I SAY\n")

		@WriteWrapper(self.wwpath, rewrite=True)
		class SomeText(Transmutable):

			def __call__(self): return "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG"
			class loggy(LibraryContrib):

				handler		= self.ACCESS_UTILS
				init_name	= "WriteWrapper_rewrite"
				init_level	= 10

		self.test_case = SomeText()

		with self.assertLogs("WriteWrapper_rewrite", 10) as case_loggy:
			self.assertEqual(self.test_case(), "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")

		self.assertTrue(os.path.isfile(self.wwpath))
		with open(self.wwpath) as ww:
			self.assertEqual(ww.read(), "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")

		self.assertIn("DEBUG:WriteWrapper_rewrite:Writing 49 symbols", case_loggy.output)
		self.assertIn(f"INFO:WriteWrapper_rewrite:Written to {self.wwpath}", case_loggy.output)




	def test_WriteWrapper_text_mode(self):

		self.fmake(self.wwpath, "I SAY\n")
		self.assertTrue(os.path.isfile(self.wwpath))
		with open(self.wwpath) as ww : self.assertEqual(ww.read(), "I SAY\n")

		@WriteWrapper(self.wwpath, text_mode=False)
		class SomeText(Transmutable):

			def __call__(self): return "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG"
			class loggy(LibraryContrib):

				handler		= self.ACCESS_UTILS
				init_name	= "WriteWrapper_text_mode"
				init_level	= 10

		self.test_case = SomeText()

		with self.assertLogs("WriteWrapper_text_mode", 10) as case_loggy:
			self.assertEqual(self.test_case(), self.wwpath)

		self.assertTrue(os.path.isfile(self.wwpath))
		with open(self.wwpath) as ww:
			self.assertEqual(ww.read(), "I SAY\nOOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")

		self.assertIn("DEBUG:WriteWrapper_text_mode:Writing 49 symbols", case_loggy.output)
		self.assertIn(f"INFO:WriteWrapper_text_mode:Written to {self.wwpath}", case_loggy.output)




	def test_WriteWrapper_bad_super(self):

		for i,bad in enumerate((

			42, .69, True, False, None, ..., print, Transmutable,
			[ "text" ],( "text", ),{ "text" },{ "text": "text" }
		)):
			with self.subTest(super=bad):

				@WriteWrapper(self.wwpath)
				class SomeText(Transmutable):

					def __call__(self): return bad
					class loggy(LibraryContrib):

						handler		= self.ACCESS_UTILS
						init_name	= f"WriteWrapper_bad_super_{i}"
						init_level	= 10

				self.test_case = SomeText()
				with self.assertLogs(f"WriteWrapper_bad_super_{i}", 10) as case_loggy:
					self.assertIsNone(self.test_case())
				self.assertIn(
					f"DEBUG:WriteWrapper_bad_super_{i}:Text to write not fetched", case_loggy.output
				)
				self.test_case.loggy.close()




	def test_WriteWrapper_bad_path(self):

		for i,bad in enumerate((

			42, .69, True, False, None, ..., print, Transmutable,
			[ "text" ],( "text", ),{ "text" },{ "text": "text" }
		)):
			with self.subTest(path=bad):

				@WriteWrapper(bad)
				class SomeText(Transmutable):

					def __call__(self): return "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG"
					class loggy(LibraryContrib):

						handler		= self.ACCESS_UTILS
						init_name	= f"WriteWrapper_bad_path_{i}"
						init_level	= 10

				self.test_case = SomeText()
				with self.assertLogs(f"WriteWrapper_bad_path_{i}", 10) as case_loggy:
					self.assertIsNone(self.test_case())
				self.assertIn(f"DEBUG:WriteWrapper_bad_path_{i}:Writing 49 symbols", case_loggy.output)
				self.assertIn(
					f"DEBUG:WriteWrapper_bad_path_{i}:Path to write is not a string", case_loggy.output
				)
				self.test_case.loggy.close()




	@unittest.skip("manual invokation only")
	def test_WriteWrapper_raise(self):

		# path argument must be a local file with unset write bits
		NO_W_PATH = ""

		@WriteWrapper(NO_W_PATH)
		class SomeText(Transmutable):

			def __call__(self): return "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG"
			class loggy(LibraryContrib):

				handler		= self.ACCESS_UTILS
				init_name	= "WriteWrapper_raise"
				init_level	= 10

		self.test_case = SomeText()

		with self.assertLogs("WriteWrapper_raise", 10) as case_loggy : self.assertIsNone(self.test_case())
		self.assertIn("DEBUG:WriteWrapper_raise:Writing 49 symbols", case_loggy.output)
		# Error message is a subject to adjust to, perhaps
		self.assertIn(

			"DEBUG:WriteWrapper_raise:Writing failed due to PermissionError: "
			f"[Errno 13] Permission denied: '{NO_W_PATH}'",
			case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from 	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.hedwig.mail.letter.fields	import SenderField
from	pygwarts.hedwig.mail.letter.fields	import RecipientField
from	pygwarts.hedwig.mail.letter.fields	import ccField
from	pygwarts.hedwig.mail.letter.fields	import bccField
from	pygwarts.hedwig.mail.letter.fields	import SubjectField
from	pygwarts.hedwig.mail.letter.fields	import BodyField
from	pygwarts.hedwig.mail.letter.fields	import AttachmentField
from	pygwarts.hedwig.mail.builder.smtp	import SMTPBuilder
from	pygwarts.hedwig.mail.utils			import EmailValidator
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access.utils			import TextWrapper
from	pygwarts.tests.hedwig				import HedwigTestCase








class SMTPCases(HedwigTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.HEDWIG_MAIL_SMTP_HANDLER): os.remove(cls.HEDWIG_MAIL_SMTP_HANDLER)

		# IMPORTANT!
		# IF CLIENT TESTS TO BE MANUALLY
		# INVOKED, BELOW REMOVES WILL NOT
		# ALLOW ATTACHMENTS TO BE INCLUDED
		if	os.path.isfile(cls.attachyone):	os.remove(cls.attachyone)
		if	os.path.isfile(cls.attachytwo):	os.remove(cls.attachytwo)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.HEDWIG_MAIL_SMTP_HANDLER)
		cls.attachyone = str(cls.HEDWIG_ROOT /"atachy.one")
		cls.attachytwo = str(cls.HEDWIG_ROOT /"atachy.two")
		if	not os.path.isfile(cls.attachyone): cls.fmake(cls, cls.attachyone, "heil maximum build!")
		if	not os.path.isfile(cls.attachytwo): cls.fmake(cls, cls.attachytwo, "heil maximum build!")

	def test_is_connectable(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "is_connectable"
				init_level	= 10

		self.test_case = CurrentBuilder()
		with self.assertLogs("is_connectable",10) as case_loggy:
			self.assertEqual(

				self.test_case.is_connectable(
					{
						"endpoint":	"emoil.comm",
						"port":		465,
						"password":	"qwerty",
					}
				),	True
			)
			self.assertEqual(len(case_loggy.output),1)
			self.assertIn("DEBUG:is_connectable:Credentials probe validated", case_loggy.output)




	def test_is_connectable_invalid(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "is_connectable"
				init_level	= 10

		self.test_case = CurrentBuilder()
		for probe in (

			"probe", 1, 1., None, True, False, ..., print, Transmutable,
			[ "endpoint", "emoil.comm", "port", 465, "password", "qwerty" ],
			( "endpoint", "emoil.comm", "port", 465, "password", "qwerty" ),
			{ "endpoint", "emoil.comm", "port", 465, "password", "qwerty" },
		):	self.assertIsNone(self.test_case.is_connectable(probe))




	def test_is_connectable_wrong(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "is_connectable_wrong"
				init_level	= 10

		self.test_case = CurrentBuilder()
		with self.assertLogs("is_connectable_wrong", 10) as case_loggy:
			self.assertEqual(
				self.test_case.is_connectable(
					{
						"not endpoint":	"emoil.com",
						"not port":		465,
						"not password":	"qwerty"
					}
				),	False
			)
		self.assertEqual(len(case_loggy.output),3)
		self.assertIn(f"INFO:is_connectable_wrong:Invalid {self.test_case} port type", case_loggy.output)
		self.assertIn(f"INFO:is_connectable_wrong:Invalid {self.test_case} endpoint type", case_loggy.output)
		self.assertIn(f"INFO:is_connectable_wrong:Invalid {self.test_case} password type", case_loggy.output)




	def test_is_connectable_fails(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "is_connectable_fails"
				init_level	= 10

		self.test_case = CurrentBuilder()
		for fail in (

			[ "fail" ],( "fail", ),{ "fail" }, { "fail": "fail" },
			None, True, False, ..., 1, 1., print, Transmutable
		):
			with self.assertLogs("is_connectable_fails", 10) as case_loggy:
				self.assertEqual(
					self.test_case.is_connectable(
						{
							"endpoint":	fail,
							"password":	"qwerty",
							"port":		465,
						}
					),	False
				)
			self.assertEqual(len(case_loggy.output),1)
			self.assertIn(
				f"INFO:is_connectable_fails:Invalid {self.test_case} endpoint type", case_loggy.output
			)

			with self.assertLogs("is_connectable_fails", 10) as case_loggy:
				self.assertEqual(
					self.test_case.is_connectable(
						{
							"password":	fail,
							"port":		465,
							"endpoint":	"endpoint",
						}
					),	False
				)
			self.assertEqual(len(case_loggy.output),1)
			self.assertIn(
				f"INFO:is_connectable_fails:Invalid {self.test_case} password type", case_loggy.output
			)

		for fail in (

			[ 465 ],( 465, ),{ 465 }, { "port": 465 },
			None, ..., "port", print, Transmutable
		):
			with self.assertLogs("is_connectable_fails", 10) as case_loggy:
				self.assertEqual(
					self.test_case.is_connectable(
						{
							"port":		fail,
							"endpoint":	"endpoint",
							"password":	"qwerty",
						}
					),	False
				)
			self.assertEqual(len(case_loggy.output),1)
			self.assertIn(f"INFO:is_connectable_fails:Invalid {self.test_case} port type", case_loggy.output)








	@unittest.skip("manual invokation only")
	def test_SMTP_build_minimal(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "SMTPBuilder_minimal"
				init_level	= 10

			class validator(EmailValidator):	pass
			# The email to send from, the first one and the only to be used
			class s1(SenderField):				field_value = "example1@email.com"
			# Second sender will not be used at all
			class s2(SenderField):				field_value = "nouse@email.com"
			class two(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

		self.test_case = CurrentBuilder()
		# Must be supplied with some working SMTP credentials to get tested
		current = self.test_case.build(
			{
				"endpoint": str(),
				"port":		int(),
				"password":	str(),
			}
		)
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)




	@unittest.skip("manual invokation only")
	def test_SMTP_build_full(self):

		class CurrentBuilder(SMTPBuilder):
			class loggy(LibraryContrib):

				handler		= self.HEDWIG_MAIL_SMTP_HANDLER
				init_name	= "SMTPBuilder_full"
				init_level	= 10

			class validator(EmailValidator):	pass
			# The email to send from, the first one and the only to be used
			class s1(SenderField):				field_value = "example1@email.com"
			# Second sender will not be used at all
			class s2(SenderField):				field_value = "exomple1@email.com"
			class r1(RecipientField):			field_value = "example2@email.com", "example3@email.com"
			class r2(RecipientField):			field_value = "example4@email.com", "example5@email.com"
			class three(ccField):				field_value = "example6@email.com", "example7@email.com"
			class four(bccField):				field_value = "example8@email.com", "example9@email.com"
			@TextWrapper("Subject: ")
			class five(SubjectField):

				field_value	= "{level} build test"
				modifiers	= { "level": "maximum" }

			@TextWrapper(header="My ", footer=", baby!")
			class six(BodyField):

				field_value	= "congratulations! Build is {quality}"
				modifiers	= { "quality": "nice" }

			class a1(AttachmentField): field_value = self.attachyone
			class a2(AttachmentField):

				field_value	= "attachytwo"
				modifiers	= lambda *S : self.attachytwo

		self.test_case = CurrentBuilder()
		# Must be supplied with some working SMTP credentials to get tested
		current = self.test_case.build(
			{
				"endpoint": str(),
				"port":		int(),
				"password":	str(),
			}
		)
		self.assertIsInstance(current, dict)
		self.assertEqual(len(current), 7)
		self.assertIn("sender", current)
		self.assertIn("recipient", current)
		self.assertIn("cc", current)
		self.assertIn("bcc", current)
		self.assertIn("subject", current)
		self.assertIn("body", current)
		self.assertIn("attachment", current)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pygwarts.tests.filch					import FilchTestCase
from	pygwarts.irma.contrib					import LibraryContrib
from	pygwarts.filch.transportclaw			import is_valid_port
from	pygwarts.filch.transportclaw			import validate_port
from	pygwarts.filch.transportclaw.scanning	import PortScanner








class TransportclawCase(FilchTestCase):

	"""
		Testing L4 instruments
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.TRNSPRTCLAW_HANDLER): os.remove(cls.TRNSPRTCLAW_HANDLER)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.TRNSPRTCLAW_HANDLER)
	def test_is_valid_port_valid(self):

		for number in range(65536):
			with self.subTest(port=number): self.assertTrue(is_valid_port(number))


	def test_is_valid_port_invalid_1(self):

		for number in range(65536):
			for port in float(number), str(number), str(float(number)):
				with self.subTest(port=port): self.assertFalse(is_valid_port(port))


	def test_is_valid_port_invalid_2(self):

		for invalid in (

			-1, -1., "-1", "-1.", 65536, 65536., "65536", "65536.",
			True, False, None, ..., print, unittest,[ 42 ],( 42, ),{ 42 },{ "port": 42 }
		):
			with self.subTest(port=invalid): self.assertFalse(is_valid_port(invalid))




	def test_validate_port_valid(self):

		for number in range(65536):
			for port in number, float(number), str(number), str(float(number)):
				with self.subTest(port=port): self.assertEqual(validate_port(port), number)


	def test_validate_port_invalid(self):
		for invalid in (

			-1, -1., "-1", "-1.", 65536, 65536., "65536", "65536.",
			True, False, None, ..., print, unittest,[ 42 ],( 42, ),{ 42 },{ "port": 42 }
		):
			self.assertIsNone(validate_port(invalid))








	def test_PortScanner(self):

		class TestScanner(PortScanner):
			class loggy(LibraryContrib):

				handler		= self.TRNSPRTCLAW_HANDLER
				init_name	= "PortScanner"
				init_level	= 10

		self.test_case = TestScanner()

		with self.assertLogs("PortScanner", 10) as case_loggy:
			current = self.test_case("127.0.0.1", 80)

		self.assertIn("DEBUG:PortScanner:Scanning 127.0.0.1:80", case_loggy.output)
		try:

			self.assertTrue(current)
			self.assertIn("INFO:PortScanner:127.0.0.1:80 listening", case_loggy.output)
		except:
			self.assertFalse(current)
			self.assertIn(

				"DEBUG:PortScanner:Socket write error %s"%(111 if os.name == "posix" else 10061),
				case_loggy.output
			)




	def test_PortScanner_invalid_ip(self):

		class TestScanner(PortScanner):
			class loggy(LibraryContrib):

				handler		= self.TRNSPRTCLAW_HANDLER
				init_name	= "PortScanner_invalid_ip"
				init_level	= 10

		self.test_case = TestScanner()
		for invalid in (

			"10.10.10.300", "LOL", 42, 69., True, False, None, ..., print, PortScanner,
			[ "10.10.10.10" ],( "10.10.10.10", ),{ "10.10.10.10" },{ "target": "10.10.10.10" }
		):
			with self.subTest(target=invalid):
				with self.assertLogs("PortScanner_invalid_ip", 10) as case_loggy:

					self.assertIsNone(self.test_case(invalid, 80))
			self.assertIn(f"DEBUG:PortScanner_invalid_ip:Invalid target \"{invalid}\"", case_loggy.output)




	def test_PortScanner_invalid_port(self):

		class TestScanner(PortScanner):
			class loggy(LibraryContrib):

				handler		= self.TRNSPRTCLAW_HANDLER
				init_name	= "PortScanner_invalid_port"
				init_level	= 10

		self.test_case = TestScanner()
		for invalid in (

			"80", "LOL", 80., True, False, None, ..., print, PortScanner,
			[ 80 ],( 80, ),{ 80 },{ "port": 80 }
		):
			with self.subTest(port=invalid):
				with self.assertLogs("PortScanner_invalid_port", 10) as case_loggy:

					self.assertIsNone(self.test_case("127.0.0.1", invalid))
			self.assertIn(f"DEBUG:PortScanner_invalid_port:Invalid port \"{invalid}\"", case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)







